/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * project_hindi
 * Copyright (C) Dilawar Singh 2010 <dilawar@ee.iitb.ac.in>
 * Copyright (C) TQ Systems GmbH (adapt to usage with Qt and playing)
 * 
 * project_hindi is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * project_hindi is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <QBuffer>
#include <QFile>
#include <QDebug>
#include <QObject>
#include <QAudioFormat>
#include <QAudioOutput>
#include <QAudioDeviceInfo>
#include <QTextStream>
#include <QtCore/QCoreApplication>

#include <stdint.h>

#include "wav-file.h"

#include	<math.h>

/* constructor */
WavFile::WavFile(QObject *parent):
            QObject(parent),
            cout(stdout)
{
    reset();
    pFile = new QFile(this);
    pWavHeader = new WAV_HDR;
    pChunkHeader = new CHUNK_HDR;
}

/* deconstructor */
WavFile::~WavFile()
{
    delete pFile;
}

void WavFile::reset()
{
    gWavDataIn.clear();
    numInSamples = 0;
    maxInSamples = 0;

    format.setFrequency(-1);
    format.setChannels(-1);
    format.setSampleSize(-1);
    format.setByteOrder(QAudioFormat::Endian(QSysInfo::ByteOrder));
    format.setSampleType(QAudioFormat::Unknown);
    format.setCodec("");
}


/*
 *--------------------------------------------------------------------------------------
 *       Class:  WavFile
 *      Method:  getNumSamples
 * Description:  returns num of sample in file.  
 *--------------------------------------------------------------------------------------
 */
long int WavFile::getNumSamples()
{
    return maxInSamples;
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  WavFile
 *      Method:  getNumChannels
 * Description:  get number of channels present in wav file. 1 is mono, 2 is
 * stereo.
 *--------------------------------------------------------------------------------------
 */
int WavFile::getNumChannels()
{
    return format.channels();
}


/*
 *--------------------------------------------------------------------------------------
 *       Class:  WavFile
 *      Method:  getBitsPerSample
 * Description:  You know. Return bit per sample.
 *--------------------------------------------------------------------------------------
 */
int WavFile::getBitsPerSample()
{
    return format.sampleSize();
}


/*
 *--------------------------------------------------------------------------------------
 *       Class:  WavFile
 *      Method:  getSampleRate
 * Description:  get sample rate in chunk.
 *--------------------------------------------------------------------------------------
 */
unsigned long WavFile::getSampleRateHz()
{
    return format.frequency();
}


/*
 *--------------------------------------------------------------------------------------
 *       Class:  WavFile
 *      Method:  ifMoreDataAvailable
 * Description:  check if something is left in file.
 *
 *--------------------------------------------------------------------------------------
 */
bool WavFile::ifMoreDataAvailable()
{
    return (numInSamples < maxInSamples);
}


/*
 *--------------------------------------------------------------------------------------
 *       Class:  WavFile
 *      Method:  readCurrentInput
 * Description:  Isn't it obvious from the file.
 *--------------------------------------------------------------------------------------
 */
int WavFile::readCurrentInput(int& Data)
{
    if( !gWavDataIn.length() || (maxInSamples <=0) || (numInSamples < 0) ) {
        qWarning("Input file not ready (or not loaded)!");
        return -1;
    }

    if( numInSamples >= maxInSamples) {
        qWarning("Nothing left to read!");
        return -2;
    }
    switch (format.sampleSize()) {
    case 16:
    {
        int16_t *pS = (int16_t*)gWavDataIn.constData();
        Data = pS[numInSamples++];
        break;
    }
    case 8:
    {
        uint8_t *pC = (uint8_t *)gWavDataIn.constData();
        Data = pC[numInSamples++];
        break;
    }
    default:
        qWarning() << "sample size " << format.sampleSize() << "not supported";
    };
    return 0;
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  WavFile
 *      Method:  WavFile
 * Description:  this is another constructor which reads the file.
 *--------------------------------------------------------------------------------------
 */
int WavFile::openWavFile(const QString& fileName)
{
    int i;

    unsigned int stat;
    char outBuffer[80];

    int sFlag;
    long int rMore;

    int wBufferLength;

    /* set the defaults values. */
    reset();

    
    if( NULL == pWavHeader )
    {
        qWarning("can't alloc new wav header\n");
        return -1;
    }

    if( NULL == pChunkHeader )
    {
        qWarning("can't alloc new chunk header\n");
        return -2;
    }

    /* 
     * open the wav file 
     */
    pFile->setFileName(fileName);
    if(!pFile->open(QIODevice::ReadOnly))
    {
        qWarning("Can't open wav file.\n");
        return -3;
    }

    /*-----------------------------------------------------------------------------
     *  Now, we have load the file. Start reading data.
     *-----------------------------------------------------------------------------*/

    /* read riff/wav header */ 
    stat = pFile->read((char*)pWavHeader, sizeof(WAV_HDR));
    if(stat != sizeof(WAV_HDR))
    {
        qWarning("Header missing. May be format is not OK!");
        return -4;
    }

    /* check format of header */
    for(i = 0; i < 4; i++)
    {
        outBuffer[i] = pWavHeader->rID[i];
    }
    outBuffer[4] = 0;
    if(strcmp(outBuffer, "RIFF") != 0)
    {
        qWarning("Bad RIFF format.");
        return -5;
    }

    for(i = 0; i < 4; i++)
    {
        outBuffer[i] = pWavHeader->wID[i];
    }
    outBuffer[4] = 0;

    if(strcmp(outBuffer, "WAVE") != 0)
    {
        qWarning("Bad WAVE format");
        return -6;
    }

    for(i = 0; i < 4; i++)
    {
        outBuffer[i] = pWavHeader->fId[i];
    }
    outBuffer[4] = 0;

    if(strcmp(outBuffer, "fmt ") != 0) // not with "fmt" since 4th pos is blank
    {
        qWarning("Bad fmt format");
        return -7;
    }

    if(pWavHeader->wFormatTag != 1)
    {
        qDebug("Bad wav wFormatTag");
        return -8;
    }

    /* 
     * Skip over any remaining portion of wav header.
     */
    rMore = pWavHeader->pcmHeaderLength - (sizeof(WAV_HDR) - 20);    
    if(!pFile->seek(pFile->pos() + rMore))
    {
        qWarning("Can't seek.");
        return -9;
    }

    /* 
     * read chunk untill a data chunk is found.
     */
    sFlag = 1;
    while(sFlag != 0)
    {
        // check attempts.
        if(sFlag > 10) { 
        	qWarning("Too many chunks"); 
        	return -10;
        }

        // read chunk header
        stat = pFile->read((char*)pChunkHeader, sizeof(CHUNK_HDR));
        if( sizeof(CHUNK_HDR) != stat)
        {
            qWarning("Reading chunk header failed");
            return -11;
        }

        // check chunk type.
        for(i =0; i < 4; i++) {
            outBuffer[i] = pChunkHeader->dId[i];
        }
        outBuffer[4] = 0;
        if(strcmp(outBuffer, "data") == 0) {
            break;
        }

        // skip over chunk.
        sFlag++;        
        if(!pFile->seek(pFile->pos() + pChunkHeader->dLen))
        {
            qWarning("Can't seek.");
            return -12;
        }
    }

    /* find length of remaining data. */
    wBufferLength = pChunkHeader->dLen;
    /* find number of samples. */
    maxInSamples = pChunkHeader->dLen;
    maxInSamples /= pWavHeader->numBitsPerSample / 8;

    /* read signal data */
    gWavDataIn = pFile->read(wBufferLength);
    if( wBufferLength != gWavDataIn.size()) {
        qWarning("Can't read buffer.");
        return -13;
    }

    /* 
     * save all this.
     */
//	newer Qt Versions support these and declare the next obsolete:
//    format.setSampleRate(getSampleRateHz());
//    format.setChannelCount(getNumChannels());
   format.setFrequency(pWavHeader->nSamplesPerSec);
   format.setChannels(pWavHeader->numChannels);
   format.setSampleSize(pWavHeader->numBitsPerSample);
   format.setCodec("audio/pcm");
   format.setByteOrder(QAudioFormat::LittleEndian);
   switch (format.sampleSize()) {
   case 8:
       format.setSampleType(QAudioFormat::UnSignedInt);
       break;
   case 16:
       format.setSampleType(QAudioFormat::SignedInt);
       break;
   default:
       qWarning() << "format not supported for playing";
       return -1;
   }

   /* reset and delete */
   numInSamples = 0;

   if(pWavHeader != NULL) delete pWavHeader;
   if(pChunkHeader != NULL) delete pChunkHeader;
   pFile->close();
    
   return 0;
}

int WavFile::displayInformation()
{
   /*
    * print the data.
    */
   cout << endl << "-----------------------------------------------------" << endl;
   cout << "Loaded wav file :" << pFile->fileName() << endl;
   cout << "Sample rate:" << format.frequency() << "(Hz)" << endl;
   cout << "Number of samples = " << maxInSamples << endl;
   cout << "Bits per sample = " << format.sampleSize() << endl;
   cout << "Number of channels = " << format.channels() << endl;
   cout << "----------------------------------------------------" << endl;
 
  return 0;
} 

const QByteArray& WavFile::getData()
{
    return gWavDataIn;
}

int WavFile::writeDataToFile(const QString& fileName)
{
    QFile *File = new QFile(fileName);
    int n;
    File->open(QIODevice::ReadOnly | QIODevice::Truncate);
    char header[50];
    n = sprintf(header, "; Sample Rate %i\n", format.frequency());
    File->write(header);
    n = sprintf(header, "; Channels %d\n", format.channels());
    File->write(header);

    for( int i = 0; i < maxInSamples; i++)
    {
       char data[30];
       int n;
       int d;
       readCurrentInput(d);
       n = sprintf(data,"\t%1.9f\t%1.9f\n", (double)i/(double)format.frequency(),
                   (double)d/ pow(2, format.sampleSize() - 1)); // normalize it.
       File->write(data);
    }
    File->close();
    return 0;
}

int WavFile::play()
{
    switch (getBitsPerSample()) {
    case 8:
        format.setSampleType(QAudioFormat::UnSignedInt);
        format.setSampleSize(8);
        break;
    case 16:
        format.setSampleType(QAudioFormat::SignedInt);
        format.setSampleSize(16);
        break;
    default:
        qWarning() << "format not supported for playing";
        return -1;
    }

    QAudioDeviceInfo info(QAudioDeviceInfo::defaultOutputDevice());
    QList<QAudioDeviceInfo> devices = QAudioDeviceInfo::availableDevices(QAudio::AudioOutput);
    int result = -88;
    for (QList<QAudioDeviceInfo>::iterator s = devices.begin(); s != devices.end(); ++s)
    {
        info = *s;
    
        qDebug() << "trying: " << info.deviceName();

        if (info.isNull()) {
            qDebug() << "invalid audio dev info";
            continue;
        }
        if (info.deviceName() == "null") {
            qDebug() << "NULL audio dev info";
            continue;
        }
        if (!info.isFormatSupported(format)) {
	        qWarning() << "Default format not supported - trying to use nearest";
	        format = info.nearestFormat(format);
		}
		QBuffer* audioBuffer = new QBuffer(&gWavDataIn, this);
		qDebug() << "audio data buffer size:" << audioBuffer->size();
		audioBuffer->open(QIODevice::ReadOnly);
        cout << "use " << info.deviceName() << " to play" << endl;
        output = new QAudioOutput(info, format, this);
        QCoreApplication::connect(output, SIGNAL(notify()), this, SLOT(onNotify()));
        QCoreApplication::connect(output, SIGNAL(stateChanged(QAudio::State)), 
        						 this, SLOT(onStateChanged(QAudio::State)));
        output->start(audioBuffer);
        
        result = 0;
        break;
    }

    return result;
}

void WavFile::start()
{
    play();
}

/* progress */
void WavFile::onNotify()
{
    if (output) {
    	cout << "bytesFree = " << output->bytesFree()
             << ", " << "elapsedUSecs = " << output->elapsedUSecs()
             << ", " << "processedUSecs = " << output->processedUSecs() << endl;
    }
}

void WavFile::onStateChanged(QAudio::State newState)
 {
     switch (newState) {
     case QAudio::StoppedState:
     	 /* external stop received or error */
         cout << "stopped state ..." << endl;
         if (output->error() != QAudio::NoError) {
             cout << "finished with error " << output->error() << endl;
         }
         emit stopped();
         break;
     case QAudio::ActiveState:
         cout << "active state ..." << endl;
         break;
     case QAudio::SuspendedState:
         cout << "suspended state ..." << endl;
         break;
     case QAudio::IdleState:
     	 /* out of data, buffer completed */
         cout << "idle state ..." << endl;
         emit stopped();
         break;
     default:
         cout << "unknown state ..." << endl;
     };
 }
 
